package com.jdbernard.wdiwtlt.cli

import com.jdbernard.wdiwtlt.ConfigWrapper
import com.jdbernard.wdiwtlt.MediaLibrary
import com.jdbernard.wdiwtlt.db.DbApi
import com.jdbernard.wdiwtlt.db.models.*
import com.jdbernard.io.NonBlockingInputStreamReader
import com.jdbernard.util.AnsiEscapeCodeSequence as ANSI
import com.zaxxer.hikari.HikariConfig
import com.zaxxer.hikari.HikariDataSource
import org.docopt.Docopt
import jline.console.ConsoleReader
import java.sql.Timestamp
import java.text.SimpleDateFormat
import java.net.URLDecoder
import java.util.regex.Pattern
import java.util.regex.Matcher
import org.slf4j.Logger
import org.slf4j.LoggerFactory
import uk.co.caprica.vlcj.discovery.NativeDiscovery
import uk.co.caprica.vlcj.component.AudioMediaListPlayerComponent
import uk.co.caprica.vlcj.player.MediaPlayerEventListener

import static com.jdbernard.util.AnsiEscapeCodeSequence.*
import static com.jdbernard.wdiwtlt.MediaLibrary.*
import static com.jdbernard.wdiwtlt.cli.CliErr.*

public class CommandLineInterface {

    public static final VERSION = "ALPHA"

    public static final def DOC = """\
wdiwtlt v$VERSION

Usage:
    wdiwtlt [options]
    wdiwtlt --version
    wdiwtlt --help

Options:
    -c --config <config-file>

        Config file for wdiwtlt CLI.

    -L --library-root <root-dir>

        The path to a local media library directory. Providing a local library
        causes wdiwtlt to run in local mode.

    -D --database-config <config-file>

        The path to a data source configuration file. wdiwtlt cli uses HikariCP
        for database connection pooling, so the given configuration file is
        expected to be formatted so that it can be fed directly to the
        HikariConfig constructor.
        (see https://github.com/brettwooldridge/HikariCP#configuration-knobs-baby)

Configuration:
"""

    private static Logger logger =
        LoggerFactory.getLogger(CommandLineInterface)

    private Properties cliConfig
    private MediaLibrary library

    /// IO Management
    private InputStream inStream
    private OutputStream outStream

    private ConsoleReader reader
    private Thread consoleReaderThread
    private List<String> consoleReadBuffer =
        Collections.synchronizedList(new ArrayList<String>())
    private synchronized boolean running

    /// Console output data
    private String titleStyle, normalStyle, statusStyle, promptStyle,
        artistStyle, albumStyle, fileStyle, errorStyle, playlistStyle,
        cmdStyle

    private String clearLine = new ANSI().eraseLine(Erase.All).toString()
    private String afterInput =
        new ANSI().eraseLine(Erase.All).scrollUp().cursorUp().toString()
    private String beforeLeader =
        new ANSI().saveCursor().cursorPrevLine(3).toString()
    private String afterLeader =
        new ANSI().restoreCursor().eraseLine(Erase.ToEnd).toString()
    private String eraseLeader =
        new ANSI().eraseLine(Erase.All).cursorPrevLine().eraseLine(Erase.All)
            .cursorPrevLine().eraseLine(Erase.All)
            .cursorPrevLine().eraseLine(Erase.All).toString()

    public final static modelClasses = [
        'album': Album, 'artist': Artist, 'bookmark': Bookmark,
        'file': MediaFile, 'mediaFile': MediaFile, 'playlist': Playlist,
        'tag': Tag ]

    public final static selectableModels = 'album|artist|file|playlist|tag'

    private int displayWidth = 79
    private long msgTimeout
    private ScrollText currentlyPlaying = new ScrollText(
        maxWidth: displayWidth - 17,
        text: "No media currently playing.")
    private ScrollText status = new ScrollText(maxWidth: displayWidth)
    private Date dismissMsgDate = new Date()
    private SimpleDateFormat sdf = new SimpleDateFormat('EEE-HH-SSS')

    /// Current play queue and selection data
    List<Model> currentSelection = []
    Playlist playQueue
    Bookmark playBookmark
    MediaFile curMediaFile

    /// VLCJ Player
    AudioMediaListPlayerComponent vlcj

    public static void main(String[] args) {

        def opts = new Docopt(DOC).withVersion("wdiwtlt v$VERSION").parse(args)

        def exitErr = { msg ->
            System.err.println("wdiwtlt: $msg")
            System.exit(1) }

        // Look for a given CLI config file.
        def givenCfg = new Properties()
        File cfgFile
        if (opts["--config"]) cfgFile = new File(opts["--config"])

        if (!cfgFile || !cfgFile.exists() || !cfgFile.isFile())
            cfgFile = new File("wdiwtlt.cli.properties")

        if (!cfgFile || !cfgFile.exists() || !cfgFile.isFile())
            cfgFile = new File(System.getenv().HOME, ".wdiwtlt.cli.properties")

        if (cfgFile.exists() && cfgFile.isFile()) {
            try { cfgFile.withInputStream { givenCfg.load(it) } }
            catch (Exception e) { givenCfg.clear() } }

        // Just in case, load up the general WDIWTLT default config service
        def wdiwtltDefaultConfig = new ConfigWrapper()

        // Find the library root directory (TODO: eventually this should be
        // enabled only when local mode is chosen).
        File libRoot = new File(
            opts["--library-root"] ?:
            givenCfg[ConfigWrapper.LIBRARY_DIR_KEY] ?:
            wdiwtltDefaultConfig.libraryRootPath ?:
            "no libRoot configured")

        if (libRoot && (!libRoot.exists() || !libRoot.isDirectory()))
            exitErr("Library root does not exist or is not a directory: " +
                libRoot.canonicalPath)

        // Get the library database
        HikariConfig hcfg
        File dbCfgFile = new File(
            opts["--database-config"] ?:
            givenCfg[ConfigWrapper.DB_CONFIG_FILE_KEY] ?:
            "database.properties")

        // Try to load from the given DB config file
        if (dbCfgFile && dbCfgFile.exists() && dbCfgFile.isFile()) {
            Properties props = new Properties()
            try { dbCfgFile.withInputStream { props.load(it) } }
            catch (Exception e) { props.clear() }

            if (props) hcfg = new HikariConfig(props) }

        // Look to see if database connection properties are given in the CLI
        // config file.
        if (givenCfg && givenCfg["database.config.dataSourceClassName"]) {
            Properties props = new Properties()
            props.putAll(givenCfg
                .findAll { it.key.startsWith("database.config.") }
                .collectEntries { [it.key[16..-1], it.value] } )
            if (props) hcfg = new HikariConfig(props) }

        // Fall back to the WDIWTLT general database defaults
        if (!hcfg) hcfg = wdiwtltDefaultConfig.hikariConfig
        if (!hcfg) exitErr("Cannot load database configuation.")

        // Create our database instance
        def dbapi
        try {
            def hds = new HikariDataSource(hcfg)
            dbapi = new DbApi(hds) }
        catch (Exception e) {
            exitErr("Could not establish a connection to the database:\n\t" +
                e.localizedMessage) }

        // Try to discover the VLC native libraries
        def vlcj
        try {
            new NativeDiscovery().discover()
            vlcj = new AudioMediaListPlayerComponent() }
        catch (Exception e) {
            exitErr("Could not find the VLC libraries. Is VLC installed?\n\t" +
                e.localizedMessage) }

        // Create our CLI object
        try {
            def cliInst = new CommandLineInterface(new MediaLibrary(dbapi, libRoot),
                vlcj, System.in, System.out, givenCfg)
            cliInst.repl() }
        catch (Exception e) { e.printStackTrace(); exitErr(e.localizedMessage) }
        finally { if (vlcj) vlcj.release() }
    }

    public CommandLineInterface(MediaLibrary library,
    AudioMediaListPlayerComponent vlcj, InputStream sin, OutputStream out,
    Properties cliConfig) {
        this.cliConfig = cliConfig
        this.library = library
        this.inStream = sin
        this.outStream = out

        this.msgTimeout = cliConfig["message.timeout"] ?: 5000l

        setupTextStyles()

        library.clean()

        playQueue = library.save(new Playlist(
            name: "CLI Queue ${sdf.format(new Date())}"))

        String user = System.getProperty('user.name')
        String os = System.getProperty('os.name')
        def bookmarks = library.getBookmarkByName("Last played for $user on $os.")
        if (bookmarks) playBookmark = bookmarks[0]
        else playBookmark = new Bookmark(name: "Last played for $user on $os.")

        playBookmark.playlistId = playQueue.id
        playBookmark.playIndex = 0
        playBookmark.mediaFileId = 0

        this.vlcj = vlcj

        def listener = new UniversalNoopImplementation()
        listener.methods.playing = this.&playing
        listener.methods.finished = this.&finished
        listener.methods.stopped = this.&stopped
        vlcj.mediaListPlayer.mediaPlayer.addMediaPlayerEventListener(
            listener as MediaPlayerEventListener)

        this.consoleReaderThread = new Thread().start(this.&runReaderThread)

    }

    void setupTextStyles() {
        titleStyle = new ANSI().color(Colors.WHITE, Colors.DEFAULT, true).toString()
        normalStyle = new ANSI().resetText().toString()
        promptStyle = new ANSI().color(Colors.YELLOW, Colors.DEFAULT, true).toString()
        cmdStyle = new ANSI().color(Colors.YELLOW, Colors.DEFAULT, true).toString()
        statusStyle = new ANSI().color(Colors.CYAN, Colors.DEFAULT, false).toString()
        playlistStyle = new ANSI().color(Colors.GREEN, Colors.DEFAULT, false).toString()
        artistStyle = new ANSI().color(Colors.RED, Colors.DEFAULT, false).toString()
        albumStyle = new ANSI().color(Colors.BLUE, Colors.DEFAULT, false).toString()
        fileStyle = new ANSI().color(Colors.GREEN, Colors.DEFAULT, false).toString()
        errorStyle = new ANSI().color(Colors.RED, Colors.DEFAULT, true).toString()
    }

    private void runReaderThread() {
        this.reader = new ConsoleReader(this.inStream, this.outStream)
        this.reader.setPrompt(promptStyle + "> " + normalStyle)

        String line
        while(this.running && !Thread.currentThread().isInterrupted()) {
            line = reader.readLine()
            if (line == null || line == "quit" || line == "exit") running = false
            else {
                consoleReadBuffer.add(line)
                outStream.print(afterInput)
                outStream.flush() } } }

    public void repl() {
        this.running = true
        String line = null
        long nextBookmarkUpdate = 0

        outStream.println "\n\n\n"
        drawLeader()

        while(this.running) {

            if (new Date() > dismissMsgDate) {
                resetStatus()
                dismissMsgDate = new Date(Long.MAX_VALUE) }

            if (consoleReadBuffer.size() > 0) {
                line = consoleReadBuffer.remove(0)
                try { processInput(line) }
                catch (CliErr cliErr) {
                    String errMsg = cliErr.message
                    logger.error(errMsg)
                    if (ANSI.strip(errMsg).length() > 80) {
                        printLongMessage(errorStyle + errMsg + normalStyle) }
                    else {
                        status.text = errorStyle + errMsg + normalStyle
                        dismissMsgDate = new Date(new Date().time + msgTimeout) } } }
            else {
                drawLeader()
                if (curMediaFile && System.currentTimeMillis() > nextBookmarkUpdate) {
                    playBookmark.playTimeMs =
                        vlcj.mediaListPlayer.mediaPlayer.time
                    library.save(playBookmark)
                    nextBookmarkUpdate = System.currentTimeMillis() + 2000 }
                Thread.sleep(250) } } }

    private def processInput(String line) {
        logger.debug("line: $line")

        String[] parts = line.split(' ', 2)
        String command = parts[0]?.toLowerCase()
        String rest = parts.size() == 2 ? parts[1]?.trim() : null
        logger.debug("command: ${command}")
        switch(command) {
            // Misc/utility
            case 'scan':        return scanMediaLibrary()
            case 'list':        return processList(rest, currentSelection)
            case 'select':      return processSelect(rest, currentSelection)
            case 'play':        return processPlay(rest, currentSelection)
            case 'enqueue':     return processEnqueue(rest, currentSelection)
            case 'add':         return processAdd(rest, currentSelection)
            case 'remove':      return processRemove(rest, currentSelection)
            case 'tag':         return processTag(rest, currentSelection)
            case 'untag':       return processTag(rest, currentSelection, false)
            case 'clear':       return processClear(rest)
            case 'pause':       return processPause()
            case 'stop':        return processStop()
            case 'n':
            case 'next':        return processNext(rest)
            case 'p':
            case 'prev':        return processPrev(rest)
            case 'jump':        return processJump(rest)
            case 'ff':
            case 'fastforward': return processFastForward(rest)
            case 'rw': case 'rwd':
            case 'rewind':      return processRewind(rest)
            case 'vol':
            case 'volume':      return processVolume(rest)
            case 'help':        return printLongMessage(processHelp(rest))

            case 'debug':
                outStream.println(
                    "\n\nConfig: \n" +
                    cliConfig.collect { "\t${it.key}: ${it.value}" }
                             .join("\n") +
                    "\n\n\n")
                drawLeader(true)
                return

            case 'q': case 'quit': case ':q': case 'exit': case '\u0004':
                running = false
                consoleReaderThread.interrupt()
                return

            default:
                err "Unrecognized command: '$line'"
                drawLeader()
                Thread.sleep(250)
                break
        }
    }

    public MediaLibrary scanMediaLibrary() {
        status.text = "Scanning media library..."
        def counts = library.rescanLibrary()
        msg("Scanned ${counts.total} files. Added ${counts.new} and " +
            "ignored ${counts.ignored} files.")
        return library }

    private String processList(String options, def selection) {
        logger.debug("Listing things. Options: $options")

        if (options == 'bookmarks') selection = library.getBookmarks()
        else if (options != 'selection') selection = select(options, selection)

        if (!selection) err "Nothing selected."
        else return printLongMessage(makeList(selection, { "${it.id}: ${it} " })) }

    private List<Model> processSelect(String options, List<Model> selection) {
        currentSelection = select(options, selection)
        if (!currentSelection) msg 'Nothing selected.'
        else resetStatus()
        logger.debug("currentSelection: $currentSelection")
        return currentSelection }

    private List<Model> select(String options, List<Model> selection = null) {

        logger.debug("Selecting: {}\tselection: {}", options, selection)
        List<Tag> excludedTags = []
        List<Tag> selectedTags = []
        Class modelClass

        switch (options) {
        case ~/playing ($selectableModels)s?/:
            if (!curMediaFile) err "No media is currently playing."

            modelClass = modelClasses[Matcher.lastMatcher[0][1]]
            logger.debug("modelClass: {}\tcurMediaFileId: {}", modelClass, curMediaFile.id)
            if (modelClass == MediaFile) return [curMediaFile]
            else if (modelClass == Playlist) return playQueue
            else return library.getWhere(modelClass,
                [mediaFileId: curMediaFile.id])

        /* TODO
        case ~/files tagged as((\s\w+)+?) and not as((\s\w+)+)/:
            excludedTags = lastMatcher[0][3].split(/\s/)
        */

        case ~/selected ($selectableModels)s?/:

            if (!selection) err "Nothing is selected."

            modelClass = modelClasses[Matcher.lastMatcher[0][1]]
            Class selectionClass = selection[0].class

            if (modelClass == selectionClass) return selection
            return selection.collectMany { library.getWhere(modelClass,
                    [(idKeyFor(selectionClass)): it.id]) }.findAll()

        case ~/files tagged( as){0,1}((\s\w+)+)/:
            selectedTags = Matcher.lastMatcher[0][2].split(/\s/)
                .collect { it?.trim() }.findAll()
                .collect { library.getTagByName(it) }.findAll()
            if (!selectedTags) err 'Nothing is selected.'
            return library.getMediaFilesWhere(tags: selectedTags)

        case ~/($selectableModels) where (.+)/:
            // TODO
            err "select <thing> where ... is not yet implemented."

        case ~/($selectableModels)s((\s\d+)+)/:
            modelClass = modelClasses[Matcher.lastMatcher[0][1]]
            return Matcher.lastMatcher[0][2].split(/\s/)
                .collect { safeToInteger(it) }.findAll()
                .collect { library.getById(modelClass, it) }.findAll()

        case ~/($selectableModels) (.+)/:
            modelClass = modelClasses[Matcher.lastMatcher[0][1]]
            String nameOrId = Matcher.lastMatcher[0][2]
            return [ensureExactlyOne(
                library.getByIdOrName(modelClass, nameOrId))]

        case ~/($selectableModels)s/:
            modelClass = modelClasses[Matcher.lastMatcher[0][1]]
            return library.getAll(modelClass)

        case 'queue': return [playQueue]
        case ~/queued ($selectableModels)s?/:
            modelClass = modelClasses[Matcher.lastMatcher[0][1]]
            return library.getWhere(modelClass, [playlistId: playQueue.id])
        case ~/untagged files/: return library.untaggedFiles
        default:
            logger.debug("Invalid select options: $options")
            err "Invalid options to the ${cmdStyle}select${normalStyle}" +
                " command. Use ${cmdStyle}help select${normalStyle} to " +
                "see a list of valid options." } }

    private Playlist processPlay(String options, List<Model> selection) {

        switch (options) {
            case null: selection = null
            case ~/selection/: break
            case ~/bookmark (.+)/:
                String nameOrId = Matcher.lastMatcher[0][1]
                Bookmark b = ensureExactlyOne(
                    library.getByIdOrName(Bookmark, nameOrId))
                if (!b) err "No bookmark matches '$nameOrId'."

                Playlist p = library.getPlaylistById(b.playlistId)
                if (!p) err 'The playlist for this bookmark no longer exists.'

                setPlayQueue(p)
                vlcj.mediaListPlayer.playItem(b.playIndex)

                if (b.playTimeMs > 0)
                    vlcj.mediaListPlayer.mediaPlayer.time = b.playTimeMs

                b.lastUsed = new Timestamp(new Date().time);
                library.update(b)

                return p;

            default: selection = select(options, selection) }

        if (selection) {
            List<MediaFile> mediaFiles = library.collectMediaFiles(selection)
            playQueue = library.removeAllFromPlaylist(playQueue.id)
            library.addToPlaylist(playQueue.id, mediaFiles.collect { it.id })
            setPlayQueue(playQueue) }

        vlcj.mediaListPlayer.play()
        return playQueue }

    private List<MediaFile> processEnqueue(String options,
    List<Model> selection = null) {
        if (options != 'selection') selection = select(options, selection)
        if (!selection) err "Nothing is selected."
        List<MediaFile> enqueued = enqueue(selection)
        msg "${enqueued.size()} files added to the current play queue."
        return enqueued }

    private List<MediaFile> enqueue(List<Model> items) {
        if (!items) return playQueue

        List<MediaFile> mediaFiles = library.collectMediaFiles(items)
        library.addToPlaylist(playQueue.id, mediaFiles.collect { it.id })
        mediaFiles.each {
            vlcj.mediaList.addMedia(
                new File(library.libraryRoot, it.filePath).canonicalPath) }

        return mediaFiles }

    private List<MediaFile> processAdd(String options,
    List<Model> selection = null) {

        Playlist p
        def m = (options =~ /(.+) to playlist (.+)/)
        p = getExactlyOne(m[0][2])
        if (!p) err "No playlist for '${Matcher.lastMatcher[0][1]}'."

        if (m[0][1] != "selection") selection = select(options, selection)

        if (!selection) err 'Nothing selected to add.'

        List<MediaFile> mediaFiles = library.collectMediaFiles(selection)
        library.addToPlaylist(p.id, mediaFiles.collect { it.id })
        msg "${mediaFiles.size()} media files added to '${p.name}'."
        return added }

    private def processRemove(String options, List<Model> selection = null) {

        def m = (options =~ /(.+) from (.+)/)
        String removeFrom = m[0][2]

        if (m[0][1] != 'selection') selection = select(m[0][1], selection)
        if (!selection) err 'Nothing was selected to be removed.'

        if (removeFrom == 'queue') p = playQueue
        else if (removeFrom.startsWith('playlist')) {
            String[] parts = removeFrom.split(/\s/, 2)
            if (parts.size() < 2) err 'No playlist id or name given.'
            p = getExactlyOne(Playlist, parts[1]) }

        List<MediaFile> toRemove = library.collectMediaFiles(selection)
        toRemove.each { library.removeFromPlaylist(p.id, it.id) }

        // Reset our queue if we removed from the queue
        if (removeFrom == 'queue') {
            vlcj.mediaListPlayer.stop()
            setPlayQueue(playQueue)

            // Restart playback with the file that was playing before we
            // removed stuff (may not be there anymore)
            if (playBookmark) {
                MediaFile mf = library.getMediaFileById(playBookmark.mediaFileId)
                List<MediaFile> playlistMFs = library.getMediaFilesWhere(
                    playlistId: playQueue.id)
                int index = playlistMFs.indexOf(mf)
                if (index > 0) {
                    vlcj.mediaListPlayer.playItem(index)
                    if (playBookmark.playTimeMs > 0) {
                        vlcj.mediaListPlayer.mediaPlayer.time =
                            playBookmark.playTimeMs } } } }

        msg "Removed ${toRemove.size()} files."
        return toRemove }

    private def processTag(String options, List<Model> selection,
    boolean addTags = true) {

        String[] parts = options.split(' as ', 2)

        List<String> tags

        // Short form: tag <tag>...
        if (parts.size() == 1) {
            if (!curMediaFile) err 'Nothing currently playing to tag.'
            selection = [curMediaFile]
            tags = parts[0].split(/\s/).collect { it?.trim() }.findAll() }

        else {
            if (parts[0] != 'selection') selection = select(parts[0], selection)
            tags = parts[1].split(/\s/).collect { it?.trim() }.findAll() }

        List<MediaFile> mediaFiles = library.collectMediaFiles(selection)

        if (addTags) library.tagMediaFiles(mediaFiles.collect { it.id }, tags)
        else library.untagMediaFiles(mediaFiles.collect {it.id}, tags)

        msg "${addTags ? 'Tagged' : 'Untagged'} ${mediaFiles.size()} files as $tags."
        return mediaFiles }

    private def processClear(String options) {

        switch(options) {
            case null:
                print(new ANSI().eraseDisplay(Erase.All))//.cursorPosition(4, 0))
                drawLeader(true)
                return
            case 'queue':
                playQueue = library.removeAllFromPlaylist(playQueue.id)
                setPlayQueue(playQueue)
                return playQueue
            case ~/playlist (.+)/:
                Playlist p = getExactlyOne(Playlist, Matcher.lastMatcher[0][1])
                if (!p) err "No playlist for '${Matcher.lastMatcher[0][1]}'."
                return library.removeAllFromPlaylist(p.id)
            case 'selection': currentSelection = []; resetStatus(); break
            default:
                err  "Unrecognized option to the ${cmdStyle}clear" +
                    "${normalStyle} command. Use ${cmdStyle}help clear" +
                    "${normalStyle} to see a list of valid options." } }

    private def processPause() { vlcj.mediaListPlayer.pause() }

    private def processStop() { vlcj.mediaListPlayer.stop() }

    private def processNext(String rest) {
        def count
        try { count = rest ? rest as int : 1 }
        catch (Exception e) { err "$count is not a valid number" }
        vlcj.mediaListPlayer.stop()
        if ((playBookmark.playIndex + count) < vlcj.mediaList.size())
            vlcj.mediaListPlayer.playItem(playBookmark.playIndex + count) }

    private def processPrev(String rest) {
        def count
        try { count = rest ? rest as int : 1 }
        catch (Exception e) { err "$count is not a valid number" }
        vlcj.mediaListPlayer.stop()
        if ((playBookmark.playIndex - count) >= 0)
            vlcj.mediaListPlayer.playItem(playBookmark.playIndex - count) }

    private def processJump(String options) {
        MediaFile target = getExactlyOne(MediaFile, options)
        if (!target) err "No media file matches '${options}'."

        int index = library.getMediaFilesWhere(playlistId: playQueue.id)
            .indexOf(target)
        if (index < 0) err "'$target' is not in the current play queue."

        vlcj.mediaListPlayer.mediaPlayer.stop()
        vlcj.mediaListPlayer.playItem(index) }

    private def processFastForward(String rest) {
        String[] parts = rest.split(' ')
        String strAmount = parts.size() > 0 ? parts[0]?.trim() : null
        String unit = parts.size() > 1 ? parts[1]?.trim() : null

        if (!strAmount) { strAmount = "10"; unit = "s" }
        if (!unit) unit = 's'

        long amount
        try { amount = strAmount as long }
        catch (Exception e) { err "$strAmount must be an integer." }

        switch (unit) {
            case 'ms': case 'millis': case 'millisecond': case 'milliseconds':
                vlcj.mediaListPlayer.mediaPlayer.skip(amount)
                break
            case 's': case 'sec': case 'second': case 'seconds':
                vlcj.mediaListPlayer.mediaPlayer.skip(amount * 1000)
                break
            case 'm': case 'min': case 'minute': case 'minutes':
                vlcj.mediaListPlayer.mediaPlayer.skip(amount * 60000)
                break
            default: err "$unit must be one of 'milliseconds' " +
                "(or 'millis' or 'ms'), 'seconds' (or 'sec' or 's'), or " +
                "'minutes' (or 'min' or 'm')" } }

    private def processRewind(String rest) {
        String[] parts = rest.split(' ')
        String strAmount = parts.size() > 0 ? parts[0]?.trim() : null
        String unit = parts.size() > 1 ? parts[1]?.trim() : null

        if (!strAmount) { strAmount = "10"; unit = "s" }
        if (!unit) unit = 's'

        long amount
        try { amount = -(strAmount as int) }
        catch (Exception e) { err "$strAmount must be an integer." }

        switch (unit) {
            case 'ms': case 'millis': case 'millisecond': case 'milliseconds':
                vlcj.mediaListPlayer.mediaPlayer.skip(amount)
                break
            case 's': case 'sec': case 'second': case 'seconds':
                vlcj.mediaListPlayer.mediaPlayer.skip(amount * 1000)
                break
            case 'm': case 'min': case 'minute': case 'minutes':
                vlcj.mediaListPlayer.mediaPlayer.skip(amount * 60000)
                break
            default: err "$unit must be one of 'milliseconds' " +
                "(or 'millis' or 'ms'), 'seconds' (or 'sec' or 's'), or " +
                "'minutes' (or 'min' or 'm')" } }

    private def processVolume(String rest) {
        int percentage

        if (rest) {
            try { percentage = Math.min(Math.max(0, rest as int), 200) }
            catch (Exception e) {
                err "Volume must be and integer between 0 and 200."
                return }
            vlcj.mediaListPlayer.mediaPlayer.volume = percentage }

        msg "Volume: ${vlcj.mediaListPlayer.mediaPlayer.volume}" }

    private String processHelp(String options) {
        switch(options) {

            // Top-level help
            // --------------
            case null: case '': return """\
Available commands:

    scan        Re-scan the media library for new files.
    list        List items (albums, artists, files, etc).
    select      Select things (into the select buffer, no the play queue).
    play        Play requested media.
    enqueue     Enqueue requested media to the end of the play queue.
    add         Append requested media to the end of a playlist.
    remove      Remove requested from the play queue or playlist.
    tag         Associate tags with requested media.
    untag       Remove the association between the given tags and media.
    clear       Clear the selection, play queue, playlist, or screen.
    pause       Pause media playback.
    stop        Stop media playback.
    next        Move forward in the play queue.
    prev        Move backward in the play queue.
    jump        Jump to a given media file in the play queue.
    ff          Jump ahead in the playback of the current media.
    rw          Jump back in the playback of the current media.
    volume      Set or retrieve the volume of the media player.
    help        Show help information about commands.
    quit        Quit this program.

A new user is advised to read the help section for the ${cmdStyle}select${normalStyle} command.

For a quick list of options, consider reading the ${cmdStyle}summary${normalStyle} section.
"""

            // SCAN
            // --------------
            case 'scan': return """\
${commandStyle}scan${normalStyle}

    Scan the media library for new files.

"""

            // LIST
            // --------------
            case 'list': return """\
${cmdStyle}list bookmarks${normalStyle}              List all bookmarks.
${cmdStyle}list selection${normalStyle}              List the currently selected items.
${cmdStyle}list <select-criteria>${normalStyle}      Make a selection using the ${cmdStyle}select${normalStyle} syntax and then
                            list the selection.

"""

            // SELECT
            // --------------
            case 'select': return """\
${cmdStyle}select playing { album | artist | file | playlist | tag }${normalStyle}

    Select the currently playing items into the selection buffer. Specifically,
    this selects items that are associated with the currently playing media
    file.

${cmdStyle}queued { albums | artists | files | playlists | tags }${normalStyle}

    Select the items currently in the queue.

${cmdStyle}select selected { album | artist | file | playlist | tag }${normalStyle}

    Select the items associated with the current selection buffer into the
    selection buffer. This is useful to change the type of the selection. For
    example, the following commands would select all of the albums in the
    library by Bob Dylan:

        select artist Bob Dylan
        select selected albums

    Another example, to select media files based on a set of tags:

        select tags instrumental orchestral
        select selected files

    This example would select all files that are tagged as *either*
    instrumental or orchestral

${cmdStyle}select files tagged as <tag-name>...${normalStyle}

    Select all media files tagged with the given tags. If multiple tags are
    given then only files which have all the given tags are selected. In
    contrast to the previous example:

        select files tagged as instrumental orchestral

    This selects all files that are tagged as *both* instrumental and
    orchestral.

${cmdStyle}select untagged files${normalStyle}

    Select all media files that do not have any tags associated with them.


${cmdStyle}select { album | artist | file | playlist | tag } where <criteria>...${normalStyle}

    ${errorStyle}Not yet implemented.${normalStyle}

${cmdStyle}select { album | artist | file | playlist | tag } <id | name>${normalStyle}

    Select a single item by ID or by name. When selecting by name, the name can
    include spaces and can be a substring of the whole name ("Lonely Hearts"
    for "Sgt. Pepper's Lonely Hearts Club Band" for example, quotations not
    required).

${cmdStyle}select { albums | artists | files | playlists | tags } <id>...${normalStyle}

    Select multiple items by ID. Multiple IDs can be given, separated by
    spaces. If no IDs are given, all of the items are returned.

${cmdStyle}queue${normalStyle}

    Select the current play queue

"""

            // PLAY
            // --------------
            case 'play': return """\
${cmdStyle}play${normalStyle}

    With no options, play the current file (inverse of pause).

${cmdStyle}play selection${normalStyle}

    Clear the play queue, enqueue the current selection, and begin playback.

${cmdStyle}play bookmark <id | name>${normalStyle}

    Load the bookmarked playlist as the play queue and begin playback at the
    bookmarked media file.

${cmdStyle}play <select-criteria>${normalStyle}

    Make a selection using the ${cmdStyle}select${normalStyle} syntax and then play the selection.

"""

            // ENQUEUE
            // --------------
            case 'enqueue': return """\
enqueue selection

    Add the media files for the selected items to the end of the current play
    queue.

enqueue <select-criteria>

    Make a selection using the ${cmdStyle}select${normalStyle} syntax and then enqueue
    the selection.

"""

            // ADD
            // --------------
            case 'add': return """\
add selection to playlist <id | name>

    Lookup a playlist by id or name (including partial match) then add the
    media files for the selected items to the end of that playlist.

add <select-criteria> to playlist <id | name>

    Lookup a playlist by id or name (including partial match), select a set of
    media files using the ${cmdStyle}select${normalStyle} syntax, then add the media files
    for the selected items to the end of that playlist.

"""

            // REMOVE
            // --------------
            case 'remove': return """\
remove selection from queue
remove selection from playlist <id | name>

    Remove the media files for the current selection from the current play
    queue or from a playlist looked up by ID or name (including partial match).

remove <select-criteria> from queue
remove <select-criteria> from playlist <id | name>

    Make a selection using the ${cmdStyle}select${normalStyle} syntax then remove those media
    files from either the play queue or from a playlist looked up by ID or name
    (including partial match).

"""

            // tag
            // --------------
            case 'tag': return """\
tag <tag>...

    Tag the currently playing file with the given tags. Multiple tags may be
    provided, separated by spaces (tags cannot include spaces).

tag selection as <tag>...

    Tag all of the media files in the current selection with the given tags.
    Multiple tags may be provided separated by spaces (tags cannot include
    spaces).

tag <select-criteria> as <tag>...

    Make a selection using the ${cmdStyle}select${normalStyle} syntax then tag all of the
    media files in the selection with the given tags. Multiple tags may be
    provided separated by spaces (tags cannot include spaces).

"""
            // UNTAG
            // --------------
            case 'untag': return """\
untag <tag>...

    Remove the given tags from the currently playing file. Multiple tags may be
    provided, separated by spaces (tags cannot include spaces).

tunag selection as <tag>...

    Remove the given tags from all of the media files in the current selection.
    Multiple tags may be provided separated by spaces (tags cannot include
    spaces).

taung <select-criteria> as <tag>...

    Make a selection using the ${cmdStyle}select${normalStyle} syntax then remove the
    given tags from all of the media files in the selection. Multiple tags may
    be provided separated by spaces (tags cannot include spaces).

"""

            // CLEAR
            // --------------
            case 'clear': """\
clear                           Clear the terminal display.
clear queue                     Clear the play queue.
clear selection                 Clear the selection buffer.
clear playlist <id | name>      Clear the given playlist

"""
            case 'pause': return 'pause     Pause playback.'
            case 'stop':  return 'stop      Stop playback'
            case 'next': return """\
next <count>        Move forward in the play queue by <count> items. <count> is
                    optional and defaults to 1

"""

            case 'prev': return """\
prev <count>        Move backward in the play queue by <count> items. <count> is
                    optional and defaults to 1

"""

            case 'jump': return """\
jump to <media file ID or name>

    Find the given media file by ID or name in the current play queue and
    resume playback starting from that file.

"""

            case 'ff': case 'fastforward': return """\
ff <amount> <unit>

    Jump forward in the playback of the current media by <amount> specified in
    <unit>s. <amount> must be an integer. <unit> may be one of: 'millisecons',
    'seconds', or 'minutes'. The following abbreviations are allowed: 'ms',
    'millis', 's', 'sec', 'm', 'min'.

"""

            case 'rw': case 'rwd': case 'rewind': return """\
rw <amount> <unit>

    Jump backward in the playback of the current media by <amount> specified in
    <unit>s. <amount> must be an integer. <unit> may be one of: 'millisecons',
    'seconds', or 'minutes'. The following abbreviations are allowed: 'ms',
    'millis', 's', 'sec', 'm', 'min'.

"""

            case 'vol': case 'volume': return """\
volume              Display the current volume setting.
volume <percent>    Set the volume. <percent> may be any value from 0 to 200.

"""

            case 'help': return """\
help summary        Display a quick summary of available commands.
help <commmand>     Display detailed information about the given command."""


            case 'summary': return """\
Selecting files:

    select { album | artist | file | playlist | tag } <id | name>
    select { albums | artists | files | playlists | tags } where <criteria>
    select files tagged as <tag>... and not as <tag>...
    select playing { albums | artists | files | playlists | tags }
    select queued { albums | artists | files | playlists | tags }

    list selection
    list <selection-criteria>
    list selected { albums | artists | files | playlists | tags }

Play and Controlling Media:

    play
    play selection
    play bookmark <id | name>
    play <selection-criteria>

    next <count>        (alias: n)
    prev <count>        (alias: p)

    stop
    pause
    fasforward <duration> <time-unit>   (aliases: ff, fwd)
    rewind <duration> <time-unit>       (aliases: rw, rwd)
    jump <id | name>

Playlist and Queue Management:

    scan
    enqueue selection
    enqueue <selection-criteria>

    add selection to playlist to playlist <id | name>
    add <selection-criteria> to playlist <id | name>

    remove selection from playlist <id | name>
    remove <selection-criteria> from playlist <id | name>
    remove selection from queue
    remove <selection-criteria> from queue

    create playlist named <name> from selection <id | name>
    create playlist named <name from <selection-criteria>
    copy playlist <id | name> as <new name>
    create bookmark <id | name>
    update bookmark <id | name>

    clear
    clear queue
    clear selection
    clear playlist <id | name>

"""

            default:
                err "Unrecognized command: '$line'"
                drawLeader()
                Thread.sleep(250)
                break } }

    private void playing(def player) {
        try {
            def mediaFiles = library.getMediaFilesWhere(playlistId: playQueue.id)
            String absFilePath =
                URLDecoder.decode(vlcj.mediaListPlayer.currentMrl()[7..-1])

            def currentIdx = mediaFiles.findIndexOf {
                def mrlPath = new File(absFilePath).canonicalPath
                def mfPath = new File(library.libraryRoot, it.filePath).canonicalPath
                return mfPath == mrlPath }

            curMediaFile = mediaFiles[currentIdx]

            currentlyPlaying.text = makeFullMediaFileDescription(curMediaFile)

            playBookmark.playlistId = playQueue.id
            playBookmark.playIndex = currentIdx
            playBookmark.mediaFileId = curMediaFile.id

            library.save(playBookmark) }
        catch (Exception e) { printLongMessage(e.printStackTrace()) } }

    private void finished(def player) {
        curMediaFile = library.incrementPlayCount(curMediaFile)
        library.save(curMediaFile) }

    private void stopped(def player) {
        curMediaFile = null
        currentlyPlaying.text = "No media currently playing." }

    public void setPlayQueue(Playlist p) {
        if (vlcj.mediaListPlayer.isPlaying()) vlcj.mediaListPlayer.stop()
        p.lastUsed = new Timestamp(new Date().time)
        playQueue = library.update(p)

        vlcj.mediaList.clear()

        library.getMediaFilesWhere(playlistId: playQueue.id).each {
            vlcj.mediaList.addMedia(
                new File(library.libraryRoot, it.filePath).canonicalPath) } }

    public def ensureExactlyOne(def matches) {
        if (!matches) err "Nothing matches."

        String englishName = toEnglish(matches[0].class)
        if (matches.size() > 1) err "Multiple  ${englishName}s match: " +
            matches.collect { "${it.id}: ${it.name}" }.join(', ')

        return matches[0] }

    public def getExactlyOne(Class modelClass, String nameOrId) {
        return ensureExactlyOne(library.getByIdOrName(modelClass, nameOrId)) }

    private void drawLeader(afterOutput = false) {

        String leader = beforeLeader + getLeader() +
            (afterOutput ?
                (promptStyle + "> " + normalStyle) :
                afterLeader)

        outStream.print(leader)
        outStream.flush() }

    private String printLongMessage(String msg) {
        String result = new StringBuilder()
            .append(eraseLeader)
            .append(msg)
            .append("\n\n\n\n")
            .toString()

        outStream.println result
        drawLeader(true)
        return result }

    private String getLeader() {
        StringBuilder leader = new StringBuilder()
            .append(clearLine)
            .append(titleStyle)
            .append("WDIWTLT - v")
            .append(VERSION)
            .append('\n')

        StringBuilder statusLine = new StringBuilder()
            .append(statusStyle)

        if (playBookmark) statusLine.append('(')
            .append(playBookmark.playIndex + 1)
            .append('/')
            .append(vlcj.mediaList.size())
            .append(') ')

        statusLine.append(currentlyPlaying)

        if (curMediaFile) {
            long playTime = Math.floor(vlcj.mediaListPlayer.mediaPlayer.time / 1000) as long
            int statusLength = ANSI.strip(statusLine.toString()).length()
            statusLine.append((' ' * (displayWidth - (statusLength + 7))))
                .append(promptStyle)
                .append(String.format('%02d', Math.floor(playTime / 60) as int))
                .append(':')
                .append(String.format('%02d', Math.floor(playTime % 60) as int))
                .append(' ')}

        leader.append(clearLine)
            .append(statusLine)
            .append("\n")
            .append(clearLine)
            .append(normalStyle)
            .append(status)
            .append("\n")

        return leader.toString() }

    private String msg(String msg) {
        status.text = msg
        dismissMsgDate = new Date(new Date().time + msgTimeout) }

    private String makeFullMediaFileDescription(MediaFile mf) {
        def artist = library.getArtistsWhere( mediaFileId: mf.id)
        def album = library.getAlbumsWhere(mediaFileId: mf.id)

        StringBuilder s = new StringBuilder()

        if (artist) s.append(artistStyle)
            .append(artist[0])
            .append(normalStyle)
            .append(" / ")

        if (album) s.append(albumStyle)
            .append(album[0])
            .append(normalStyle)
            .append(" / ")

        s.append(fileStyle)
         .append(mf)
         .append(normalStyle)
         .append(' ')

        return s.toString() }

    private String makeList(List<Model> items, Closure toString = null) {

        if (!items) return "No items to list."

        if (!toString) toString = { it.toString() }

        Class modelClass = items[0].class
        def currentCollection

        if (curMediaFile) {
            if (modelClass == MediaFile) currentCollection = [curMediaFile]
            else currentCollection = library.getWhere(
                modelClass, [mediaFileId: curMediaFile.id]) }

        def highlightSelected = { item ->
            if (currentCollection && currentCollection.contains(item))
                return "${promptStyle}${toString(item)}${normalStyle}"
            else return toString(item) }

        def result = new StringBuilder()
            .append("--------------------\n${modelClass.simpleName}s:\n\n")

        result.append(items.collect(highlightSelected).join("\n"))
            .append("\n")

        return result.toString() }

    private String resetStatus() {

        if (currentSelection) {
            if (currentSelection.size() == 1) {
                String s
                Model m = currentSelection[0]
                switch (m.class) {
                    case Album: s = "$albumStyle$m"; break
                    case Artist: s = "$artistStyle$m"; break
                    case Playlist: s = "$playlistStyle$m"; break
                    case MediaFile:
                        s = makeFullMediaFileDescription(currentSelection[0])
                        break
                    case Tag: s = "Tag: $m" }

                status.text = s }
            else status.text = "${currentSelection.size()} " +
            "${toEnglish(currentSelection[0].class)}s selected." }
        return status.text }

}
